/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _GitHubActionsCache_instances, _GitHubActionsCache_baseUrl, _GitHubActionsCache_authToken, _GitHubActionsCache_logger, _GitHubActionsCache_hitRateLimit, _GitHubActionsCache_reserveUploadAndCommitTarball, _GitHubActionsCache_upload, _GitHubActionsCache_commit, _GitHubActionsCache_request, _GitHubActionsCache_onRateLimit, _GitHubActionsCache_computeCacheKey, _GitHubActionsCache_computeVersion, _GitHubActionsCache_makeTarball, _GitHubActionsCache_reserveCacheEntry, _GitHubActionsCacheHit_instances, _GitHubActionsCacheHit_script, _GitHubActionsCacheHit_url, _GitHubActionsCacheHit_applied, _GitHubActionsCacheHit_download, _GitHubActionsCacheHit_extract;
import * as pathlib from 'path';
import * as fs from 'fs/promises';
import * as https from 'https';
import { createHash } from 'crypto';
import { scriptReferenceToString } from '../script.js';
import { getScriptDataDir } from '../util/script-data-dir.js';
import { execFile } from 'child_process';
import { createReadStream, createWriteStream } from 'fs';
/**
 * Caches script output to the GitHub Actions caching service.
 */
export class GitHubActionsCache {
    constructor(logger, baseUrl, authToken) {
        _GitHubActionsCache_instances.add(this);
        _GitHubActionsCache_baseUrl.set(this, void 0);
        _GitHubActionsCache_authToken.set(this, void 0);
        _GitHubActionsCache_logger.set(this, void 0);
        /**
         * Once we've hit a 429 rate limit error from GitHub, simply stop hitting the
         * cache for the remainder of this Wireit process. Caching is not critical,
         * it's just an optimization.
         *
         * TODO(aomarks) We could be a little smarter and do retries, but this at
         * least should stop builds breaking in the short-term.
         */
        _GitHubActionsCache_hitRateLimit.set(this, false);
        __classPrivateFieldSet(this, _GitHubActionsCache_baseUrl, baseUrl, "f");
        __classPrivateFieldSet(this, _GitHubActionsCache_authToken, authToken, "f");
        __classPrivateFieldSet(this, _GitHubActionsCache_logger, logger, "f");
    }
    static create(logger) {
        // The ACTIONS_CACHE_URL and ACTIONS_RUNTIME_TOKEN environment variables are
        // automatically provided to GitHub Actions re-usable workflows. However,
        // they are _not_ provided to regular "run" scripts. For this reason, we
        // re-export those variables so that all "run" scripts can access them using
        // the "google/wireit@setup-github-actions-caching/v1" re-usable workflow.
        //
        // https://github.com/actions/toolkit/blob/500d0b42fee2552ae9eeb5933091fe2fbf14e72d/packages/cache/src/internal/cacheHttpClient.ts#L38
        const baseUrl = process.env['ACTIONS_CACHE_URL'];
        if (!baseUrl) {
            return {
                ok: false,
                error: {
                    reason: 'invalid-usage',
                    message: 'The ACTIONS_CACHE_URL variable was not set, but is required when ' +
                        'WIREIT_CACHE=github. Use the google/wireit@setup-github-cache/v1 ' +
                        'action to automatically set environment variables.',
                },
            };
        }
        if (!baseUrl.endsWith('/')) {
            // Internally, the @actions/cache library expects the URL to end with a
            // slash. While we could be more lenient, we want to match the behavior of
            // any other calls happening inside that library which we don't control.
            return {
                ok: false,
                error: {
                    reason: 'invalid-usage',
                    message: `The ACTIONS_CACHE_URL must end in a forward-slash, got ${JSON.stringify(baseUrl)}.`,
                },
            };
        }
        // https://github.com/actions/toolkit/blob/500d0b42fee2552ae9eeb5933091fe2fbf14e72d/packages/cache/src/internal/cacheHttpClient.ts#L63
        const authToken = process.env['ACTIONS_RUNTIME_TOKEN'];
        if (!authToken) {
            return {
                ok: false,
                error: {
                    reason: 'invalid-usage',
                    message: 'The ACTIONS_RUNTIME_TOKEN variable was not set, but is required when ' +
                        'WIREIT_CACHE=github. Use the google/wireit@setup-github-cache/v1 ' +
                        'action to automatically set environment variables.',
                },
            };
        }
        return {
            ok: true,
            value: new GitHubActionsCache(logger, baseUrl, authToken),
        };
    }
    async get(script, fingerprint) {
        if (__classPrivateFieldGet(this, _GitHubActionsCache_hitRateLimit, "f")) {
            return undefined;
        }
        const version = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeVersion).call(this, fingerprint);
        const key = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeCacheKey).call(this, script);
        const url = new URL('_apis/artifactcache/cache', __classPrivateFieldGet(this, _GitHubActionsCache_baseUrl, "f"));
        url.searchParams.set('keys', key);
        url.searchParams.set('version', version);
        const { req, resPromise } = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_request).call(this, url);
        req.end();
        const res = await resPromise;
        if (res.statusCode === /* No Content */ 204) {
            return undefined;
        }
        if (isOk(res)) {
            const { archiveLocation } = JSON.parse(await readBody(res));
            return new GitHubActionsCacheHit(script, archiveLocation);
        }
        if (res.statusCode === /* Too Many Requests */ 429) {
            __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_onRateLimit).call(this, script);
            return;
        }
        throw new Error(`GitHub Cache check HTTP ${String(res.statusCode)} error: ` +
            (await readBody(res)));
    }
    async set(script, fingerprint, relativeFiles) {
        if (__classPrivateFieldGet(this, _GitHubActionsCache_hitRateLimit, "f")) {
            return false;
        }
        const absFiles = relativeFiles.map((rel) => pathlib.join(script.packageDir, rel.path));
        const tempDir = await makeTempDir(script);
        try {
            const tarballPath = await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_makeTarball).call(this, [...absFiles], tempDir);
            return await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_reserveUploadAndCommitTarball).call(this, script, fingerprint, tarballPath);
        }
        finally {
            await fs.rm(tempDir, { recursive: true });
        }
    }
}
_GitHubActionsCache_baseUrl = new WeakMap(), _GitHubActionsCache_authToken = new WeakMap(), _GitHubActionsCache_logger = new WeakMap(), _GitHubActionsCache_hitRateLimit = new WeakMap(), _GitHubActionsCache_instances = new WeakSet(), _GitHubActionsCache_reserveUploadAndCommitTarball = 
/**
 * @returns True if we reserved, uploaded, and committed the tarball. False if
 * we gave up due to a rate limit error.
 * @throws If an unexpected HTTP error occured.
 */
async function _GitHubActionsCache_reserveUploadAndCommitTarball(script, fingerprint, tarballPath) {
    const tarballStats = await fs.stat(tarballPath);
    const tarballBytes = tarballStats.size;
    // Reference:
    // https://github.com/actions/toolkit/blob/f8a69bc473af4a204d0c03de61d5c9d1300dfb17/packages/cache/src/cache.ts#L174
    const GB = 1024 * 1024 * 1024;
    const maxBytes = 10 * GB;
    if (tarballBytes > maxBytes) {
        __classPrivateFieldGet(this, _GitHubActionsCache_logger, "f").log({
            script,
            type: 'info',
            detail: 'generic',
            message: `Output was too big to be cached: ` +
                `${Math.round(tarballBytes / GB)}GB > ` +
                `${Math.round(maxBytes / GB)}GB.`,
        });
        return false;
    }
    const id = await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_reserveCacheEntry).call(this, script, __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeCacheKey).call(this, script), __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeVersion).call(this, fingerprint), tarballBytes);
    // It's likely that we'll occasionally fail to reserve an entry and get
    // undefined here, especially when running multiple GitHub Action jobs in
    // parallel with the same scripts, because there is a window of time between
    // calling "get" and "set" on the cache in which another worker could have
    // reserved the entry before us. Non fatal, just don't save.
    if (id === undefined) {
        return false;
    }
    if (!(await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_upload).call(this, script, id, tarballPath, tarballBytes))) {
        return false;
    }
    if (!(await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_commit).call(this, script, id, tarballBytes))) {
        return false;
    }
    return true;
}, _GitHubActionsCache_upload = 
/**
 * @returns True if we uploaded, false if we gave up due to a rate limit error.
 * @throws If an unexpected HTTP error occured.
 */
async function _GitHubActionsCache_upload(script, id, tarballPath, tarballBytes) {
    const url = new URL(`_apis/artifactcache/caches/${id}`, __classPrivateFieldGet(this, _GitHubActionsCache_baseUrl, "f"));
    // Reference:
    // https://github.com/actions/toolkit/blob/500d0b42fee2552ae9eeb5933091fe2fbf14e72d/packages/cache/src/options.ts#L59
    const maxChunkSize = 32 * 1024 * 1024;
    const tarballHandle = await fs.open(tarballPath, 'r');
    let offset = 0;
    try {
        // TODO(aomarks) Chunks could be uploaded in parallel.
        while (offset < tarballBytes) {
            const chunkSize = Math.min(tarballBytes - offset, maxChunkSize);
            const start = offset;
            const end = offset + chunkSize - 1;
            offset += maxChunkSize;
            const tarballChunkStream = createReadStream(tarballPath, {
                fd: tarballHandle.fd,
                start,
                end,
                autoClose: false,
            });
            const opts = {
                method: 'PATCH',
                headers: {
                    'content-type': 'application/octet-stream',
                    'content-range': `bytes ${start}-${end}/*`,
                },
            };
            const { req, resPromise } = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_request).call(this, url, opts);
            tarballChunkStream.pipe(req);
            tarballChunkStream.on('close', () => {
                req.end();
            });
            const res = await resPromise;
            if (res.statusCode === /* Too Many Requests */ 429) {
                __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_onRateLimit).call(this, script);
                return false;
            }
            if (!isOk(res)) {
                throw new Error(`GitHub Cache upload HTTP ${String(res.statusCode)} error: ${await readBody(res)}\nopts: ${JSON.stringify(opts)}`);
            }
        }
        return true;
    }
    finally {
        await tarballHandle.close();
    }
}, _GitHubActionsCache_commit = 
/**
 * @returns True if we committed, false if we gave up due to a rate limit error.
 * @throws If an unexpected HTTP error occured.
 */
async function _GitHubActionsCache_commit(script, id, tarballBytes) {
    const url = new URL(`_apis/artifactcache/caches/${String(id)}`, __classPrivateFieldGet(this, _GitHubActionsCache_baseUrl, "f"));
    const reqBody = JSON.stringify({
        size: tarballBytes,
    });
    const { req, resPromise } = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_request).call(this, url, {
        method: 'POST',
        headers: {
            'content-type': 'application/json',
        },
    });
    req.end(reqBody);
    const res = await resPromise;
    if (res.statusCode === /* Too Many Requests */ 429) {
        __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_onRateLimit).call(this, script);
        return false;
    }
    if (!isOk(res)) {
        throw new Error(`GitHub Cache commit HTTP ${String(res.statusCode)} error: ${await readBody(res)}`);
    }
    return true;
}, _GitHubActionsCache_request = function _GitHubActionsCache_request(url, options) {
    return request(url, {
        ...options,
        headers: {
            // https://github.com/actions/toolkit/blob/500d0b42fee2552ae9eeb5933091fe2fbf14e72d/packages/cache/src/internal/cacheHttpClient.ts#L55
            accept: 'application/json;api-version=6.0-preview.1',
            // https://github.com/actions/toolkit/blob/500d0b42fee2552ae9eeb5933091fe2fbf14e72d/packages/http-client/src/auth.ts#L46
            authorization: `Bearer ${__classPrivateFieldGet(this, _GitHubActionsCache_authToken, "f")}`,
            ...options?.headers,
        },
    });
}, _GitHubActionsCache_onRateLimit = function _GitHubActionsCache_onRateLimit(script) {
    if (__classPrivateFieldGet(this, _GitHubActionsCache_hitRateLimit, "f")) {
        return;
    }
    __classPrivateFieldGet(this, _GitHubActionsCache_logger, "f").log({
        script,
        type: 'info',
        detail: 'generic',
        message: `Hit GitHub Actions cache rate limit, caching disabled.`,
    });
    __classPrivateFieldSet(this, _GitHubActionsCache_hitRateLimit, true, "f");
}, _GitHubActionsCache_computeCacheKey = function _GitHubActionsCache_computeCacheKey(script) {
    return createHash('sha256')
        .update(scriptReferenceToString(script))
        .digest('hex');
}, _GitHubActionsCache_computeVersion = function _GitHubActionsCache_computeVersion(fingerprint) {
    const parts = [
        fingerprint.string,
        'gzip',
        // The ImageOS environment variable tells us which operating system
        // version is being used for the worker VM (e.g. "ubuntu20",
        // "macos11"). We already include process.platform in the fingerprint,
        // but this is more specific.
        //
        // There is also an ImageVersion variable (e.g. "20220405.4") which we
        // could consider including, but it probably changes frequently and is
        // unlikely to affect output, so we prefer the higher cache hit rate.
        process.env.ImageOS ?? '',
        // Versioning salt:
        //   - <omitted>: Initial version.
        //   - 2: Removed empty directories manifest.
        '2',
    ];
    return createHash('sha256')
        .update(parts.join('\x1E') // ASCII record seperator
    )
        .digest('hex');
}, _GitHubActionsCache_makeTarball = 
/**
 * Create a tarball file in a local temp directory containing the given paths.
 *
 * @returns The full path to the tarball file on disk.
 */
async function _GitHubActionsCache_makeTarball(paths, tempDir) {
    // Create a manifest file so that we can pass a large number of files to
    // tar.
    const manifestPath = pathlib.join(tempDir, 'manifest.txt');
    await fs.writeFile(manifestPath, paths.join('\n'), 'utf8');
    const tarballPath = pathlib.join(tempDir, 'cache.tgz');
    await new Promise((resolve, reject) => {
        execFile('tar', [
            // Use the newer standardized tar format.
            '--posix',
            // Use gzip compression.
            //
            // TODO(aomarks) zstd is faster and has better performance, but it's
            // availability is unreliable, and appears to have a bug on Windows
            // (https://github.com/actions/cache/issues/301). Investigate and
            // enable if easy.
            '--gzip',
            '--create',
            '--file',
            tarballPath,
            // Use absolute paths (note we use the short form because the long
            // form is --absolute-names on GNU tar, but --absolute-paths on BSD
            // tar).
            '-P',
            // We have a complete list of files and directories, so we don't need
            // or want tar to automatically expand directories. This also allows
            // us to create empty directories, even if they aren't actually empty
            // on disk.
            '--no-recursion',
            '--files-from',
            manifestPath,
        ], (error) => {
            if (error != null) {
                reject(`tar error: ${String(error)}`);
            }
            else {
                resolve();
            }
        });
    });
    return tarballPath;
}, _GitHubActionsCache_reserveCacheEntry = 
/**
 * Reserve a cache entry.
 *
 * @returns A numeric cache id the cache entry was reserved for us, or
 * undefined if the cache entry was already reserved, or a rate limit error
 * occured.
 */
async function _GitHubActionsCache_reserveCacheEntry(script, key, version, cacheSize) {
    const url = new URL('_apis/artifactcache/caches', __classPrivateFieldGet(this, _GitHubActionsCache_baseUrl, "f"));
    const reqBody = JSON.stringify({
        key,
        version,
        cacheSize,
    });
    const { req, resPromise } = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_request).call(this, url, {
        method: 'POST',
        headers: {
            'content-type': 'application/json',
        },
    });
    req.end(reqBody);
    const res = await resPromise;
    if (isOk(res)) {
        const resData = JSON.parse(await readBody(res));
        return resData.cacheId;
    }
    if (res.statusCode === /* Conflict */ 409) {
        return undefined;
    }
    if (res.statusCode === /* Too Many Requests */ 429) {
        __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_onRateLimit).call(this, script);
        return undefined;
    }
    throw new Error(`GitHub Cache reserve HTTP ${String(res.statusCode)} error: ${await readBody(res)}`);
};
class GitHubActionsCacheHit {
    constructor(script, location) {
        _GitHubActionsCacheHit_instances.add(this);
        _GitHubActionsCacheHit_script.set(this, void 0);
        _GitHubActionsCacheHit_url.set(this, void 0);
        _GitHubActionsCacheHit_applied.set(this, false);
        __classPrivateFieldSet(this, _GitHubActionsCacheHit_script, script, "f");
        __classPrivateFieldSet(this, _GitHubActionsCacheHit_url, location, "f");
    }
    async apply() {
        if (__classPrivateFieldGet(this, _GitHubActionsCacheHit_applied, "f")) {
            throw new Error('GitHubActionsCacheHit.apply was called more than once');
        }
        __classPrivateFieldSet(this, _GitHubActionsCacheHit_applied, true, "f");
        const tempDir = await makeTempDir(__classPrivateFieldGet(this, _GitHubActionsCacheHit_script, "f"));
        const tarballPath = pathlib.join(tempDir, 'cache.tgz');
        try {
            // TODO(aomarks) Recover from rate limits and other HTTP errors.
            await __classPrivateFieldGet(this, _GitHubActionsCacheHit_instances, "m", _GitHubActionsCacheHit_download).call(this, tarballPath);
            await __classPrivateFieldGet(this, _GitHubActionsCacheHit_instances, "m", _GitHubActionsCacheHit_extract).call(this, tarballPath);
        }
        finally {
            await fs.rm(tempDir, { recursive: true });
        }
    }
}
_GitHubActionsCacheHit_script = new WeakMap(), _GitHubActionsCacheHit_url = new WeakMap(), _GitHubActionsCacheHit_applied = new WeakMap(), _GitHubActionsCacheHit_instances = new WeakSet(), _GitHubActionsCacheHit_download = async function _GitHubActionsCacheHit_download(tarballPath) {
    const { req, resPromise } = request(__classPrivateFieldGet(this, _GitHubActionsCacheHit_url, "f"));
    req.end();
    const res = await resPromise;
    if (!isOk(res)) {
        throw new Error(`GitHub Cache download HTTP ${String(res.statusCode)} error`);
    }
    await new Promise((resolve, reject) => {
        const writeTarballStream = createWriteStream(tarballPath);
        writeTarballStream.on('error', (error) => reject(error));
        res.on('error', (error) => reject(error));
        res.pipe(writeTarballStream);
        writeTarballStream.on('close', () => {
            resolve();
        });
    });
}, _GitHubActionsCacheHit_extract = function _GitHubActionsCacheHit_extract(tarballPath) {
    return new Promise((resolve, reject) => {
        execFile('tar', ['--extract', '--file', tarballPath, '--gzip', '-P'], (error) => {
            if (error != null) {
                reject(`tar error: ${String(error)}`);
            }
            else {
                resolve();
            }
        });
    });
};
function request(url, options) {
    const opts = {
        ...options,
        headers: {
            // https://github.com/actions/toolkit/blob/500d0b42fee2552ae9eeb5933091fe2fbf14e72d/packages/cache/src/internal/cacheHttpClient.ts#L67
            'user-agent': 'actions/cache',
            ...options?.headers,
        },
    };
    let req;
    const resPromise = new Promise((resolve, reject) => {
        req = https.request(url, opts, (res) => {
            resolve(res);
        });
        req.on('error', (error) => {
            reject(error);
        });
    });
    return { req, resPromise };
}
function isOk(res) {
    return (res.statusCode !== undefined &&
        res.statusCode >= 200 &&
        res.statusCode < 300);
}
function readBody(res) {
    const chunks = [];
    res.on('data', (chunk) => {
        chunks.push(chunk);
    });
    return new Promise((resolve, reject) => {
        res.on('error', (error) => {
            reject(error);
        });
        res.on('end', () => {
            resolve(Buffer.concat(chunks).toString());
        });
    });
}
function makeTempDir(script) {
    return fs.mkdtemp(pathlib.join(getScriptDataDir(script), 'temp'));
}
//# sourceMappingURL=github-actions-cache.js.map