/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
import { shuffle } from '../util/shuffle.js';
/**
 * A single execution of a specific script.
 */
export class BaseExecution {
    constructor(script, executor, logger) {
        Object.defineProperty(this, "script", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "executor", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "logger", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.script = script;
        this.executor = executor;
        this.logger = logger;
    }
    /**
     * Execute all of this script's dependencies.
     */
    async executeDependencies() {
        // Randomize the order we execute dependencies to make it less likely for a
        // user to inadvertently depend on any specific order, which could indicate
        // a missing edge in the dependency graph.
        shuffle(this.script.dependencies);
        const dependencyResults = await Promise.all(this.script.dependencies.map((dependency) => {
            return this.executor.execute(dependency.config);
        }));
        const results = [];
        const errors = new Set();
        for (let i = 0; i < dependencyResults.length; i++) {
            const result = dependencyResults[i];
            if (!result.ok) {
                for (const error of result.error) {
                    errors.add(error);
                }
            }
            else {
                results.push([this.script.dependencies[i].config, result.value]);
            }
        }
        if (errors.size > 0) {
            return { ok: false, error: [...errors] };
        }
        return { ok: true, value: results };
    }
}
//# sourceMappingURL=base.js.map