/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Executor_instances, _Executor_executions, _Executor_logger, _Executor_workerPool, _Executor_cache, _Executor_failureOccured, _Executor_stopStartingNewScripts, _Executor_killRunningScripts, _Executor_executeAccordingToKind;
import { NoOpExecution } from './execution/no-op.js';
import { OneShotExecution } from './execution/one-shot.js';
import { scriptReferenceToString } from './script.js';
import { Deferred } from './util/deferred.js';
import { convertExceptionToFailure } from './error.js';
/**
 * Executes a script that has been analyzed and validated by the Analyzer.
 */
export class Executor {
    constructor(logger, workerPool, cache, failureMode, abort) {
        _Executor_instances.add(this);
        _Executor_executions.set(this, new Map());
        _Executor_logger.set(this, void 0);
        _Executor_workerPool.set(this, void 0);
        _Executor_cache.set(this, void 0);
        /** Resolves when the first failure occurs in any script. */
        _Executor_failureOccured.set(this, new Deferred());
        /** Resolves when we decide that new scripts should not be started. */
        _Executor_stopStartingNewScripts.set(this, new Deferred());
        /** Resolves when we decide that running scripts should be killed. */
        _Executor_killRunningScripts.set(this, new Deferred());
        __classPrivateFieldSet(this, _Executor_logger, logger, "f");
        __classPrivateFieldSet(this, _Executor_workerPool, workerPool, "f");
        __classPrivateFieldSet(this, _Executor_cache, cache, "f");
        // If this entire execution is aborted because e.g. the user sent a SIGINT
        // to the Wireit process, then dont start new scripts, and kill running
        // ones.
        void abort.promise.then(() => {
            __classPrivateFieldGet(this, _Executor_stopStartingNewScripts, "f").resolve();
            __classPrivateFieldGet(this, _Executor_killRunningScripts, "f").resolve();
        });
        // If a failure occurs, then whether we stop starting new scripts or kill
        // running ones depends on the failure mode setting.
        void __classPrivateFieldGet(this, _Executor_failureOccured, "f").promise.then(() => {
            switch (failureMode) {
                case 'continue': {
                    break;
                }
                case 'no-new': {
                    __classPrivateFieldGet(this, _Executor_stopStartingNewScripts, "f").resolve();
                    break;
                }
                case 'kill': {
                    __classPrivateFieldGet(this, _Executor_stopStartingNewScripts, "f").resolve();
                    __classPrivateFieldGet(this, _Executor_killRunningScripts, "f").resolve();
                    break;
                }
                default: {
                    const never = failureMode;
                    throw new Error(`Internal error: unexpected failure mode: ${String(never)}`);
                }
            }
        });
    }
    /**
     * Signal that a script has failed, which will potentially stop starting or
     * kill other scripts depending on the {@link FailureMode}.
     *
     * This method will be called automatically in the normal flow of execution,
     * but scripts can also call it directly to synchronously signal a failure.
     */
    notifyFailure() {
        __classPrivateFieldGet(this, _Executor_failureOccured, "f").resolve();
    }
    /**
     * Synchronously check if new scripts should stop being started.
     */
    get shouldStopStartingNewScripts() {
        return __classPrivateFieldGet(this, _Executor_stopStartingNewScripts, "f").settled;
    }
    /**
     * A promise which resolves if we should kill running scripts.
     */
    get shouldKillRunningScripts() {
        return __classPrivateFieldGet(this, _Executor_killRunningScripts, "f").promise;
    }
    async execute(script) {
        const executionKey = scriptReferenceToString(script);
        let promise = __classPrivateFieldGet(this, _Executor_executions, "f").get(executionKey);
        if (promise === undefined) {
            promise = __classPrivateFieldGet(this, _Executor_instances, "m", _Executor_executeAccordingToKind).call(this, script)
                .catch((error) => convertExceptionToFailure(error, script))
                .then((result) => {
                if (!result.ok) {
                    this.notifyFailure();
                }
                return result;
            });
            __classPrivateFieldGet(this, _Executor_executions, "f").set(executionKey, promise);
        }
        return promise;
    }
}
_Executor_executions = new WeakMap(), _Executor_logger = new WeakMap(), _Executor_workerPool = new WeakMap(), _Executor_cache = new WeakMap(), _Executor_failureOccured = new WeakMap(), _Executor_stopStartingNewScripts = new WeakMap(), _Executor_killRunningScripts = new WeakMap(), _Executor_instances = new WeakSet(), _Executor_executeAccordingToKind = function _Executor_executeAccordingToKind(script) {
    if (script.command === undefined) {
        return NoOpExecution.execute(script, this, __classPrivateFieldGet(this, _Executor_logger, "f"));
    }
    return OneShotExecution.execute(script, this, __classPrivateFieldGet(this, _Executor_workerPool, "f"), __classPrivateFieldGet(this, _Executor_cache, "f"), __classPrivateFieldGet(this, _Executor_logger, "f"));
};
//# sourceMappingURL=executor.js.map