/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Fingerprint_str, _Fingerprint_data;
import { createHash } from 'crypto';
import { createReadStream } from 'fs';
import { resolve } from 'path';
import { glob } from './util/glob.js';
import { scriptReferenceToString } from './script.js';
/**
 * The fingerprint of a script. Converts lazily between string and data object
 * forms.
 */
export class Fingerprint {
    constructor() {
        _Fingerprint_str.set(this, void 0);
        _Fingerprint_data.set(this, void 0);
    }
    static fromString(string) {
        const fingerprint = new Fingerprint();
        __classPrivateFieldSet(fingerprint, _Fingerprint_str, string, "f");
        return fingerprint;
    }
    /**
     * Generate the fingerprint data object for a script based on its current
     * configuration, input files, and the fingerprints of its dependencies.
     */
    static async compute(script, dependencyFingerprints) {
        let allDependenciesAreCacheable = true;
        const filteredDependencyFingerprints = [];
        for (const [dep, depFingerprint] of dependencyFingerprints) {
            if (!depFingerprint.data.cacheable) {
                allDependenciesAreCacheable = false;
            }
            filteredDependencyFingerprints.push([
                scriptReferenceToString(dep),
                depFingerprint.data,
            ]);
        }
        let fileHashes;
        if (script.files?.values.length) {
            const files = await glob(script.files.values, {
                cwd: script.packageDir,
                absolute: false,
                followSymlinks: true,
                // TODO(aomarks) This means that empty directories are not reflected in
                // the fingerprint, however an empty directory could modify the behavior
                // of a script. We should probably include empty directories; we'll just
                // need special handling when we compute the fingerprint, because there
                // is no hash we can compute.
                includeDirectories: false,
                // We must expand directories here, because we need the complete
                // explicit list of files to hash.
                expandDirectories: true,
                throwIfOutsideCwd: false,
                stats: false,
            });
            // TODO(aomarks) Instead of reading and hashing every input file on every
            // build, use inode/mtime/ctime/size metadata (which is much faster to
            // read) as a heuristic to detect files that have likely changed, and
            // otherwise re-use cached hashes that we store in e.g.
            // ".wireit/<script>/hashes".
            fileHashes = await Promise.all(files.map(async (file) => {
                const absolutePath = resolve(script.packageDir, file.path);
                const hash = createHash('sha256');
                for await (const chunk of createReadStream(absolutePath)) {
                    hash.update(chunk);
                }
                return [file.path, hash.digest('hex')];
            }));
        }
        else {
            fileHashes = [];
        }
        const cacheable = 
        // If command is undefined, then we simply propagate the fingerprints of
        // our dependencies, and don't have any effect ourselves on cacheability.
        script.command === undefined ||
            // If input files are undefined, then it's not safe to be fresh or cached,
            // because we wouldn't know when an input file has changed that could
            // affect the output.
            (script.files !== undefined &&
                // If output files are undefined, then it's not safe to be fresh,
                // because we aren't able to check whether the output files were
                // independently modified since the last run. It's also not possible to
                // be cached, because we wouldn't know what to put in the cache.
                script.output !== undefined &&
                // If any of our dependencies are uncacheable, then we're uncacheable
                // too, because that dependency could be producing an output that is an
                // untracked input of ours.
                allDependenciesAreCacheable);
        const fingerprint = new Fingerprint();
        // Note: The order of all fields is important so that we can do fast string
        // comparison.
        const data = {
            cacheable,
            platform: process.platform,
            arch: process.arch,
            nodeVersion: process.version,
            command: script.command?.value,
            extraArgs: script.extraArgs ?? [],
            clean: script.clean,
            files: Object.fromEntries(fileHashes.sort(([aFile], [bFile]) => aFile.localeCompare(bFile))),
            output: script.output?.values ?? [],
            dependencies: Object.fromEntries(filteredDependencyFingerprints.sort(([aRef], [bRef]) => aRef.localeCompare(bRef))),
        };
        __classPrivateFieldSet(fingerprint, _Fingerprint_data, data, "f");
        return fingerprint;
    }
    get string() {
        if (__classPrivateFieldGet(this, _Fingerprint_str, "f") === undefined) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            __classPrivateFieldSet(this, _Fingerprint_str, JSON.stringify(__classPrivateFieldGet(this, _Fingerprint_data, "f")), "f");
        }
        return __classPrivateFieldGet(this, _Fingerprint_str, "f");
    }
    get data() {
        if (__classPrivateFieldGet(this, _Fingerprint_data, "f") === undefined) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            __classPrivateFieldSet(this, _Fingerprint_data, JSON.parse(__classPrivateFieldGet(this, _Fingerprint_str, "f")), "f");
        }
        return __classPrivateFieldGet(this, _Fingerprint_data, "f");
    }
    equal(other) {
        return this.string === other.string;
    }
}
_Fingerprint_str = new WeakMap(), _Fingerprint_data = new WeakMap();
//# sourceMappingURL=fingerprint.js.map