/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _IdeAnalyzer_instances, _IdeAnalyzer_overlayFs, _IdeAnalyzer_analyzer, _IdeAnalyzer_getInfoAboutLocation;
import * as fs from 'fs/promises';
import { Analyzer } from './analyzer.js';
import * as url from 'url';
import * as pathlib from 'path';
import * as jsonParser from 'jsonc-parser';
import { offsetInsideRange, OffsetToPositionConverter, } from './error.js';
class OverlayFilesystem {
    constructor() {
        // filename to contents
        Object.defineProperty(this, "overlay", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
    }
    async readFile(path, options) {
        const contents = this.overlay.get(path);
        if (contents !== undefined) {
            return contents;
        }
        return fs.readFile(path, options);
    }
}
/**
 * The interface for an IDE to communicate with wireit's analysis pipeline.
 *
 * An IDE has certain files open with in-memory buffers. These buffers often
 * shadow the files on disk, and in those cases we want to use the buffer if
 * it's available, and fall back on disk contents if not.
 *
 * Generally the user only cares about the in-memory files, at least for
 * most features like diagnostics.
 */
export class IdeAnalyzer {
    constructor() {
        _IdeAnalyzer_instances.add(this);
        _IdeAnalyzer_overlayFs.set(this, void 0);
        _IdeAnalyzer_analyzer.set(this, void 0);
        __classPrivateFieldSet(this, _IdeAnalyzer_overlayFs, new OverlayFilesystem(), "f");
        __classPrivateFieldSet(this, _IdeAnalyzer_analyzer, new Analyzer(__classPrivateFieldGet(this, _IdeAnalyzer_overlayFs, "f")), "f");
    }
    /**
     * Adds the file to the set of open files if it wasn't already,
     * and specifies its contents. Open files are defined by their
     * in memory contents, not by their on-disk contents.
     *
     * We also only care about diagnostics for open files.
     *
     * IDEs will typically call this method when a user opens a package.json file
     * for editing, as well as once for each edit the user makes.
     */
    setOpenFileContents(path, contents) {
        __classPrivateFieldGet(this, _IdeAnalyzer_overlayFs, "f").overlay.set(path, contents);
        __classPrivateFieldSet(this, _IdeAnalyzer_analyzer, new Analyzer(__classPrivateFieldGet(this, _IdeAnalyzer_overlayFs, "f")), "f");
    }
    /**
     * Removes a file from the set of open files.
     */
    closeFile(path) {
        __classPrivateFieldGet(this, _IdeAnalyzer_overlayFs, "f").overlay.delete(path);
        __classPrivateFieldSet(this, _IdeAnalyzer_analyzer, new Analyzer(__classPrivateFieldGet(this, _IdeAnalyzer_overlayFs, "f")), "f");
    }
    get openFiles() {
        return __classPrivateFieldGet(this, _IdeAnalyzer_overlayFs, "f").overlay.keys();
    }
    /**
     * Calculates and returns diagnostics for open files. If a file has no
     * diagnostics then we don't include an entry for it at all.
     */
    async getDiagnostics() {
        const diagnostics = new Map();
        function addDiagnostic(diagnostic) {
            const path = diagnostic.location.file.path;
            if (!openFiles.has(path)) {
                return;
            }
            const converted = convertDiagnostic(diagnostic);
            let set = diagnostics.get(path);
            if (set === undefined) {
                set = new Set();
                diagnostics.set(path, set);
            }
            set.add(converted);
        }
        const openFiles = new Set(this.openFiles);
        for (const failure of await __classPrivateFieldGet(this, _IdeAnalyzer_analyzer, "f").analyzeFiles([...openFiles])) {
            if (failure.diagnostic != null) {
                addDiagnostic(failure.diagnostic);
            }
            if (failure.diagnostics != null) {
                for (const diagnostic of failure.diagnostics) {
                    addDiagnostic(diagnostic);
                }
            }
        }
        return diagnostics;
    }
    async getCodeActions(path, range) {
        const codeActions = [];
        // file isn't open
        if (!__classPrivateFieldGet(this, _IdeAnalyzer_overlayFs, "f").overlay.has(path)) {
            return codeActions;
        }
        const packageDir = pathlib.dirname(path);
        // If there are any syntax-level errors for the file, we don't want to
        // offer any code actions.
        const packageJsonResult = await __classPrivateFieldGet(this, _IdeAnalyzer_analyzer, "f").getPackageJson(packageDir);
        if (!packageJsonResult.ok || packageJsonResult.value.failures.length > 0) {
            return codeActions;
        }
        const packageJson = packageJsonResult.value;
        const ourRange = OffsetToPositionConverter.get(packageJson.jsonFile).ideRangeToRange(range);
        const scriptInfo = await __classPrivateFieldGet(this, _IdeAnalyzer_instances, "m", _IdeAnalyzer_getInfoAboutLocation).call(this, packageJson, ourRange.offset);
        if (scriptInfo === undefined) {
            return codeActions;
        }
        if (scriptInfo.kind === 'dependency') {
            // No code actions for dependencies yet.
            return codeActions;
        }
        const { script, scriptSyntaxInfo: { name, scriptNode, wireitConfigNode }, } = scriptInfo;
        if (scriptInfo.kind === 'scripts-section-script' &&
            scriptNode != null &&
            wireitConfigNode == null) {
            const edit = getEdit(packageJson.jsonFile, [
                { path: ['scripts', name], value: 'wireit' },
                {
                    path: ['wireit', name],
                    value: { command: scriptNode.value },
                },
            ]);
            codeActions.push({
                title: `Refactor this script to use wireit.`,
                kind: 'refactor.extract',
                edit,
            });
        }
        if (scriptInfo.kind === 'wireit-section-script' && scriptNode == null) {
            const edit = getEdit(packageJson.jsonFile, [
                { path: ['scripts', script.name], value: 'wireit' },
            ]);
            codeActions.push({
                title: `Add this script to the "scripts" section.`,
                /**
                 * Quoting https://microsoft.github.io//language-server-protocol/specifications/lsp/3.17/specification/
                 *
                 * > 'Fix all' actions automatically fix errors that have a clear fix
                 * > that do not require user input. They should not suppress errors
                 * > or perform unsafe fixes such as generating new types or classes.
                 */
                kind: 'source.fixAll',
                edit,
            });
        }
        if (scriptNode == null ||
            wireitConfigNode == null ||
            scriptNode.value === 'wireit') {
            return codeActions;
        }
        // Ok, so there's definitely a wireit config and an entry in the scripts
        // section, however the scripts section has its own command.
        // In this case, we want to offer the user the option to move that command
        // into the wireit section, but we need to be careful that we don't
        // lose the user's command.
        // Let's find the command, if any, in the wireit config.
        let wireitCommand;
        for (const propNode of wireitConfigNode.children ?? []) {
            if (propNode.type !== 'property') {
                continue;
            }
            const [key, value] = propNode.children ?? [];
            if (key?.value !== 'command') {
                continue;
            }
            if (typeof value.value !== 'string') {
                return codeActions; // This is invalid, so we don't offer anything.
            }
            wireitCommand = value.value;
            break;
        }
        if (wireitCommand === undefined) {
            // this is the easy case, we can just move the command over
            const edit = getEdit(packageJson.jsonFile, [
                { path: ['scripts', script.name], value: 'wireit' },
                { path: ['wireit', script.name, 'command'], value: scriptNode.value },
            ]);
            codeActions.push({
                title: `Move this script's command into the wireit config.`,
                // This is mostly safe. The user might have moved the command
                // back to the scripts section because they don't want some wireit
                // features, like they don't want it to clean, or to run dependencies.
                // So we don't want it to happen automatically, but it's very safe.
                kind: 'quickfix',
                isPreferred: true,
                edit,
            });
            return codeActions;
        }
        // In the case where the commands are the same, we can just replace the
        // script version with "wireit"
        if (wireitCommand === scriptNode.value) {
            const edit = getEdit(packageJson.jsonFile, [
                { path: ['scripts', script.name], value: 'wireit' },
            ]);
            codeActions.push({
                title: `Run "wireit" in the scripts section.`,
                kind: 'quickfix',
                isPreferred: true,
                edit,
            });
            return codeActions;
        }
        // Ok here's the tricky part, we could lose user data.
        const edit = getEdit(packageJson.jsonFile, [
            { path: ['scripts', script.name], value: 'wireit' },
            {
                path: ['wireit', script.name, '[the script command was]'],
                value: scriptNode.value,
            },
        ]);
        codeActions.push({
            title: `Move this script's command into the wireit config.`,
            kind: 'quickfix',
            isPreferred: false,
            edit,
        });
        return codeActions;
    }
    async getDefinition(path, position) {
        const packageDir = pathlib.dirname(path);
        const packageJsonResult = await __classPrivateFieldGet(this, _IdeAnalyzer_analyzer, "f").getPackageJson(packageDir);
        if (!packageJsonResult.ok) {
            return undefined;
        }
        const packageJson = packageJsonResult.value;
        const ourPosition = OffsetToPositionConverter.get(packageJson.jsonFile).idePositionToOffset(position);
        const scriptInfo = await __classPrivateFieldGet(this, _IdeAnalyzer_instances, "m", _IdeAnalyzer_getInfoAboutLocation).call(this, packageJson, ourPosition);
        if (scriptInfo?.kind === 'dependency') {
            const dep = scriptInfo.dependency;
            const targetFile = dep.config.declaringFile;
            const targetNode = dep.config.configAstNode ?? dep.config.scriptAstNode;
            if (targetFile == null || targetNode == null) {
                return;
            }
            const targetConverter = OffsetToPositionConverter.get(targetFile);
            const sourceConverter = OffsetToPositionConverter.get(packageJson.jsonFile);
            return [
                {
                    originSelectionRange: sourceConverter.toIdeRange(scriptInfo.dependency.astNode),
                    targetUri: url.pathToFileURL(targetFile.path).toString(),
                    targetRange: targetConverter.toIdeRange(
                    // The parent is the property, including both key and value.
                    // So we preview the whole thing when looking at the definition:
                    //      "build": {"command": "tsc"}
                    //      ~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    targetNode.parent ?? targetNode),
                    targetSelectionRange: targetConverter.toIdeRange(targetNode.name),
                },
            ];
        }
        if (scriptInfo?.kind === 'scripts-section-script') {
            const sourceConverter = OffsetToPositionConverter.get(packageJson.jsonFile);
            const syntaxInfo = scriptInfo.scriptSyntaxInfo;
            if (syntaxInfo.scriptNode && syntaxInfo.wireitConfigNode) {
                // we can jump from the script section to the wireit config
                return [
                    {
                        originSelectionRange: sourceConverter.toIdeRange(syntaxInfo.scriptNode.parent ?? syntaxInfo.scriptNode),
                        targetUri: url.pathToFileURL(packageJson.jsonFile.path).toString(),
                        targetRange: sourceConverter.toIdeRange(syntaxInfo.wireitConfigNode.parent ?? syntaxInfo.wireitConfigNode),
                        targetSelectionRange: sourceConverter.toIdeRange(syntaxInfo.wireitConfigNode.name),
                    },
                ];
            }
        }
    }
    async getPackageJsonForTest(filename) {
        const packageDir = pathlib.dirname(filename);
        const packageJsonResult = await __classPrivateFieldGet(this, _IdeAnalyzer_analyzer, "f").getPackageJson(packageDir);
        if (!packageJsonResult.ok) {
            return undefined;
        }
        return packageJsonResult.value;
    }
}
_IdeAnalyzer_overlayFs = new WeakMap(), _IdeAnalyzer_analyzer = new WeakMap(), _IdeAnalyzer_instances = new WeakSet(), _IdeAnalyzer_getInfoAboutLocation = async function _IdeAnalyzer_getInfoAboutLocation(packageJson, offset) {
    const locationInfo = packageJson.getInfoAboutLocation(offset);
    if (locationInfo === undefined) {
        return;
    }
    const script = await __classPrivateFieldGet(this, _IdeAnalyzer_analyzer, "f").analyzeIgnoringErrors({
        name: locationInfo.scriptSyntaxInfo.name,
        packageDir: pathlib.dirname(packageJson.jsonFile.path),
    });
    for (const dep of script.dependencies ?? []) {
        if (offsetInsideRange(offset, dep.astNode)) {
            return {
                kind: 'dependency',
                dependency: dep,
                script: script,
                scriptSyntax: locationInfo.scriptSyntaxInfo,
            };
        }
    }
    return {
        ...locationInfo,
        script,
    };
};
function getEdit(file, modifications) {
    const edits = [];
    for (const { path, value } of modifications) {
        edits.push(...jsonParser.modify(file.contents, path, value, inferModificationOptions(file)));
    }
    const converter = OffsetToPositionConverter.get(file);
    const textEdits = edits.map((e) => {
        return {
            range: converter.toIdeRange(e),
            newText: e.content,
        };
    });
    return { changes: { [file.path]: textEdits } };
}
function inferModificationOptions(file) {
    const firstPostNewlineWhitespace = file.contents.match(/\n(\s+)/)?.[1];
    if (firstPostNewlineWhitespace === undefined) {
        return {};
    }
    if (/^ +$/.test(firstPostNewlineWhitespace)) {
        return {
            formattingOptions: {
                insertSpaces: true,
                tabSize: firstPostNewlineWhitespace.length,
            },
        };
    }
    else if (/^\t+$/.test(firstPostNewlineWhitespace)) {
        return {
            formattingOptions: {
                insertSpaces: false,
                tabSize: firstPostNewlineWhitespace.length,
            },
        };
    }
    return {};
}
function convertDiagnostic(d) {
    const converter = OffsetToPositionConverter.get(d.location.file);
    let relatedInformation;
    if (d.supplementalLocations) {
        relatedInformation = [];
        for (const loc of d.supplementalLocations) {
            relatedInformation.push({
                location: {
                    uri: url.pathToFileURL(loc.location.file.path).toString(),
                    range: converter.toIdeRange(loc.location.range),
                },
                message: loc.message,
            });
        }
    }
    return {
        severity: convertSeverity(d.severity),
        message: d.message,
        source: 'wireit',
        range: converter.toIdeRange(d.location.range),
        relatedInformation,
    };
}
function convertSeverity(severity) {
    switch (severity) {
        case 'error':
            return 1; // DiagnosticSeverity.Error;
        case 'warning':
            return 2; // DiagnosticSeverity.Warning;
        case 'info':
            return 3; // DiagnosticSeverity.Information;
        default: {
            const never = severity;
            throw new Error(`Unexpected severity: ${String(never)}`);
        }
    }
}
//# sourceMappingURL=ide.js.map